unit Postage;
  (* From a given stock of stamps and an amount of necessary postage, pick
  sets of stamps to minimize number of stamps and excess postage *)

  interface
    uses
      AdsCnnct (* tAdsConnection *),
      AdsData  (* used by AdsFunc *),
      AdsFunc  (* used by AdsTable *),
      AdsTable (* tAdsQuery, tAdsTable *),
      AdvSpin  (* tAdvSpinEdit *),
      Buttons  (* tBitBtn *),
      Classes  (**),
      Controls (* crHourglass, tCursor *),
      DB       (* tDataSource *),
      DBCtrls  (* tDBEdit, tDBNavigator *),
      DBGrids  (* tDBGrid *),
      ExtCtrls (* tRadioGroup; also used by AdvSpin *),
      Forms    (* Application, tForm, Screen *),
      Grids    (* used by DBGrids *),
      Mask     (* used by DBCtrls *),
      MyUtils  (* tRRFCloseBtn *),
      StdCtrls (* tButton, tCheckBox, tListBox *),
      Windows;

    type
        tfmChooseStamps = class(tForm)
          cnHousehold: tAdsConnection;
            qGetSmallestStamp: tAdsQuery;
            qGetSolution:      tAdsQuery;
            qGetTotalsUsed:    tAdsQuery;
            qSetNUsed:         tAdsQuery;
            qStamps:           tAdsQuery;
              dataStamps: tDataSource;
                dbeditComment: tDBEdit;
                dbgridStamps: tDBGrid;
                navStamps: tDBNavigator;
            qZeroStampsUsed:   tAdsQuery;
          spinPostageNeeded: tAdvSpinEdit;
          btnStop: tBitBtn;
          btnGo: tButton;
          ckbxDeductible: tCheckBox;
          ckbxVerbose:    tCheckBox;
          lbxSolutions: tListBox;
          rgCountry: tRadioGroup;
          btnClose: tRRFCloseBtn;
        procedure qStampsNewRecord(DataSet: tDataSet);

        procedure btnGoClick  (Sender: tObject);
        procedure btnStopClick(Sender: tObject);
        procedure FormShow    (Sender: tObject);
        procedure UIChanged   (Sender: tObject);
      private
        StopWanted: boolean;
        Country: string;
        procedure DoIt;
        procedure SetNUsed(const ThisStamp: currency; const N: integer);
        procedure ShowStamps;
      end (* tfmChooseStamps *);

    var
      fmChooseStamps: tfmChooseStamps;
      (* eject *)
  implementation
    uses
      MyConsts (* nullstring *),
      SysUtils (* Format, IntToStr, Trim *);
      {$R *.dfm}

    procedure tfmChooseStamps.btnStopClick(Sender: tObject);
      (* set the flag *)
    
      begin (* tfmChooseStamps.btnStopClick *)
        StopWanted := true
      end   (* tfmChooseStamps.btnStopClick *);

    procedure tfmChooseStamps.btnGoClick(Sender: tObject);
      (* fill the grid with candidates *)

      var
        OldCursor: tCursor;

      begin (* tfmChooseStamps.btnGoClick *)
        OldCursor := Screen.Cursor;
        Screen.Cursor := crHourglass;
        DoIt;
        Screen.Cursor := OldCursor
      end   (* tfmChooseStamps.btnGoClick *);

    procedure tfmChooseStamps.DoIt;
      (* guts of the program *)

      var
        Solved: boolean;
        Amount, BestN, BestTotal, NAtBestTotal, TotalAtBestN: integer;

      procedure UseSmallestStampBiggerThan(const BiggestStampUsed: currency;
                                           const NStampsSoFar, TotalSoFar: integer;
                                           Indent: string = '   ');
        (* greedy algorithm *)

        var
          ThisStamp: currency;
          CentsNeeded, Denomination, MaxUsable, NewTotal, NewNStamps, NUsed,
            OnHand: integer;

        procedure CheckForSolution;
          (* report a solution iff
                sum(NUsed * ThisDenomination) >= Amount and
                sum(Nused) < BestN  or  sum(NUsed * ThisDenomination) < BestTotal
          *)

          var
            LooksGood: boolean;
            Denomination: currency;
            NStamps, NUsed, Total: integer;
            TheLine: string;

          begin (* CheckForSolution *)
            with qGetTotalsUsed do
              begin (* qGetTotalsUsed *)
                Close;
                Open;
                NStamps := FieldByName('NStamps').AsInteger;
                Total   := FieldByName('Total'  ).AsInteger;
                Close
              end   (* qGetTotalsUsed *);

            if Total >= Amount then
              begin (* nondegenerate *)
                LooksGood := false;
                if (NStamps < BestN) or
                   ((NStamps = BestN) and (Total < TotalAtBestN)) then
                  begin (* better N *)
                    BestN := NStamps;
                    TotalAtBestN := Total;
                    LooksGood := true
                  end   (* better N *);

                if (Total < BestTotal) or
                   ((Total = BestTotal) and (Nstamps < NAtBestTotal)) then
                  begin (* fewer stamps *)
                    BestTotal := Total;
                    NAtBestTotal := NStamps;
                    LooksGood := true
                  end   (* fewer stamps *);

                if LooksGood then
                  with qGetSolution do
                    begin (* qGetSolution *)
                      TheLine := Format('%s totalling $%.2f:',
                                        [Plural(NStamps, 'stamp'),
                                         Total * 0.01]);
                      Close;
                      Open;
                      while not EOF do
                        begin (* one record *)
                          Denomination := FieldByName('Denomination').AsCurrency;
                          NUsed        := FieldByName('NUsed'       ).AsInteger;
                          TheLine := Format('%s %d x %.2m,',
                                            [TheLine, NUsed, Denomination]);
                          Next
                        end   (* one record *);

                      TheLine := copy(TheLine, 1, length(TheLine)-1);
                      with lbxSolutions do
                        ItemIndex := Items.Add(TheLine);
                      Solved := true;
                      Close
                    end   (* qGetSolution *)
              end   (* nondegenerate *)
          end   (* CheckForSolution *);

        begin (* UseSmallestStampBiggerThan *)
          Application.ProcessMessages;
          if not StopWanted then
            with qGetSmallestStamp do
              begin (* qGetSmallestStamp *)
                Close;
                ParamByName('BiggestStampUsed').AsCurrency := BiggestStampUsed;
                ParamByName('Country'         ).AsString  := Country;
                ParamByName('Deductible'      ).AsBoolean  := ckbxDeductible.Checked;
                Open;

                ThisStamp := FieldByName('Denomination').AsCurrency;
                OnHand    := FieldByName('OnHand'      ).AsInteger;
                if OnHand > 0 then
                  begin (* nondegenerate *)
                    CentsNeeded := Amount - TotalSoFar;
                    Denomination := trunc(100 * (ThisStamp + 0.009));
                    MaxUsable := (CentsNeeded + Denomination - 1) div Denomination;
                    if MaxUsable > OnHand then
                      MaxUsable := OnHand;

                    for NUsed := 0 to MaxUsable do
                      begin (* try using NUsed of these stamps *)
                        if ckbxVerbose.Checked then
                          begin (* show details *)
                            with lbxSolutions do
                              ItemIndex :=
                                Items.Add(Format('%susing %d of %d on top of %d making %d',
                                                 [Indent, NUsed, Denomination, NStampsSoFar, TotalSoFar]));
                            lbxSolutions.Repaint
                          end   (* show details *);

                        NewTotal := TotalSoFar + NUsed * Denomination;
                        NewNStamps := NStampsSoFar + NUsed;

                        if ((NewNStamps > BestN) and (NewTotal > BestTotal)) or
                           ((BestTotal = Amount) and (NewNStamps >= NAtBestTotal)) then
                          break

                        else
                          begin (* not hopeless *)
                            SetNUsed(ThisStamp, NUsed);

                            if NewTotal >= Amount then
                              CheckForSolution
                            else
                              UseSmallestStampBiggerThan(ThisStamp, NewNStamps, NewTotal, Indent + '   ')
                          end   (* not hopeless *)
                      end   (* try using NUsed of these stamps *);

                    SetNUsed(ThisStamp, 0)
                  end   (* nondegenerate *)
              end   (* qGetSmallestStamp *)
        end   (* UseSmallestStampBiggerThan *);

      begin (* tfmChooseStamps.DoIt *)
        BestN := maxint;
        BestTotal := maxint;
        NAtBestTotal := maxint;
        TotalAtBestN := maxint;
        Amount := trunc(100 * (spinPostageNeeded.FloatValue + 0.009));
        lbxSolutions.Clear;
        lbxSolutions.Repaint;
        qZeroStampsUsed.ExecSQL;
        Solved := false;
        StopWanted := false;

        if Amount > 0 then
          UseSmallestStampBiggerThan(0, 0, 0);

        if not Solved then
          with lbxSolutions do
            ItemIndex := Items.Add('(need more stamps)')
      end   (* tfmChooseStamps.DoIt *);

    procedure tfmChooseStamps.FormShow(Sender: tObject);
      (* Load the list of stamps on hand from the registry *)

      begin (* tfmChooseStamps.FormShow *)
        UIChanged(nil)
      end   (* tfmChooseStamps.FormShow *);

    procedure tfmChooseStamps.qStampsNewRecord(DataSet: tDataSet);
      (* copy from UI *)

      begin (* tfmChooseStamps.qStampsNewRecord *)
        with DataSet do
          begin (* DataSet *)
            FieldByName('Country'   ).AsString  := Country;
            FieldByName('Deductible').AsBoolean := ckbxDeductible.Checked
          end   (* DataSet *)
      end   (* tfmChooseStamps.qStampsNewRecord *);

    procedure tfmChooseStamps.SetNUsed(const ThisStamp: currency;
                                       const N: integer);
      (* say that N stamps of this denomination are in use *)

      begin (* tfmChooseStamps.SetNUsed *)
        with qSetNUsed do
          begin (* qSetNUsed *)
            ParamByName('Country'     ).AsString   := Country;
            ParamByName('Deductible'  ).AsBoolean  := ckbxDeductible.Checked;
            ParamByName('Denomination').AsCurrency := ThisStamp;
            ParamByName('NUsed'       ).AsInteger  := N;
            ExecSQL
          end   (* qSetNUsed *)
      end   (* tfmChooseStamps.SetNUsed *);

    procedure tfmChooseStamps.ShowStamps;
      (* make the query take account of the UI *)

      begin (* tfmChooseStamps.ShowStamps *)
        with qStamps do
          begin (* qStamps *)
            Close;
            ParamByName('Country'   ).AsString  := Country;
            ParamByName('Deductible').AsBoolean := ckbxDeductible.Checked;
            Open
          end   (* qStamps *)
      end   (* tfmChooseStamps.ShowStamps *);

    procedure tfmChooseStamps.UIChanged(Sender: tObject);
      (* filter table differently *)

      begin (* tfmChooseStamps.UIChanged *)
        with rgCountry do
          Country := Items[ItemIndex];
        ShowStamps
      end   (* tfmChooseStamps.UIChanged *);

end.
